#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <graphics.h>
#include <dos.h>
#include <mem.h>
#include <math.h>

#define REC_SIZE	18
#define HEX_SIZE	22
#define TILE_REC	0
#define TILE_HEX	1

//---------------------------------------------------------------------------
// represent a cell in the mine grid
//---------------------------------------------------------------------------
typedef struct cell
{
	int	ismine;
	int	cleared;
	int	marked;
	int	count;
} cell;

int game_over = 0;
int game_exit = 0;
int game_tile = 0;
int game_font = 0;
int grid_w = 0;
int grid_h = 0;

cell grid [50][50];

int init_grid ();
int draw_grid ();
int show_mouse ();
int hide_mouse ();


//---------------------------------------------------------------------------
// print text to screen
//---------------------------------------------------------------------------
int print_text (int x, int y, char* str, int color = RED, int size = 1)
{
	settextstyle (DEFAULT_FONT, HORIZ_DIR, size);
	settextjustify (CENTER_TEXT, CENTER_TEXT);
	setcolor (color);
	outtextxy (x, y, str);
	return (0);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
int show_mouse ()
{
	union REGS in, out;
	in.x.ax = 1;
	int86 (51, &in, &out);
	return (0);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------

int hide_mouse ()
{
	union REGS in, out;
	in.x.ax = 2;
	int86 (51, &in, &out);
	return (0);
}

//---------------------------------------------------------------------------
// get the mouse status: left, right button...
//---------------------------------------------------------------------------
int mouse_status (int& x, int& y, int& click)
{
	union REGS in, out;
	in.x.ax = 3;
	int86 (51, &in, &out);

	click = out.x.bx;
	x = out.x.cx;
	y = out.x.dx;
	return (0);
}

//---------------------------------------------------------------------------
// limit the mouse cursor postion
//---------------------------------------------------------------------------
int limit_mouse ()
{
	union REGS in, out;
	in.x.ax = 7;
	in.x.cx = 0;
	in.x.dx = getmaxx ();
	int86 (51, &in, &out);

	in.x.ax = 8;
	in.x.cx = 0;
	in.x.dx = getmaxy ();
	int86 (51, &in, &out);

	return (0);
}

//---------------------------------------------------------------------------
// init_game
//---------------------------------------------------------------------------
int init_game ()
{
	// init graphics, using the default EGAVGA 640x480x16
	int gdriver = DETECT, gmode, errorcode;
	initgraph (&gdriver, &gmode, "");
	errorcode = graphresult();
	if (errorcode != grOk)
	{
		printf ("Graphics error: %s\n", grapherrormsg(errorcode));
		printf ("Press any key to halt:");
		getch ();
		exit (1);
	}
	game_over = 0;
	init_grid ();
	draw_grid ();
	return (0);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
int term_game ()
{
	hide_mouse ();
	closegraph ();
	return (0);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
int valid_index (int x, int y)
{
	if (x >= 0 && x < grid_w && y >=0 && y < grid_h) return (1);
	return (0);
}

//---------------------------------------------------------------------------
//
//---------------------------------------------------------------------------
int init_grid ()
{
	if (game_tile == TILE_REC)
	{
		grid_w = (getmaxx () + 1) / REC_SIZE;
		grid_h = (getmaxy () - 19) / REC_SIZE;
	}else
	{
		grid_w = (getmaxx () + 1) / (HEX_SIZE * 1.5);
		grid_h = (getmaxy () - 19) / HEX_SIZE * 2;
	}
	int i, j, k, h;
	randomize ();
	for (i = 0; i < grid_w; i ++)
	for (j = 0; j < grid_h; j ++)
	{
		memset (&grid[i][j], 0, sizeof(cell));
		int r = random (100);
		if (r < 20)
		{
			grid [i][j].ismine = 1;
		}
	}
	for (i = 0; i < grid_w; i ++)
	for (j = 0; j < grid_h; j ++)
	{
		if (game_tile == TILE_REC)
		{
			if (1 == valid_index (i - 1, j - 1))
				grid [i][j].count += grid [i - 1][j - 1].ismine;
			if (1 == valid_index (i, j - 1))
				grid [i][j].count += grid [i][j - 1].ismine;
			if (1 == valid_index (i + 1, j - 1))
				grid [i][j].count += grid [i + 1][j - 1].ismine;
			if (1 == valid_index (i - 1, j))
				grid [i][j].count += grid [i - 1][j].ismine;
			if (1 == valid_index (i + 1, j))
				grid [i][j].count += grid [i + 1][j].ismine;
			if (1 == valid_index (i - 1, j + 1))
				grid [i][j].count += grid [i - 1][j + 1].ismine;
			if (1 == valid_index (i, j + 1))
				grid [i][j].count += grid [i][j + 1].ismine;
			if (1 == valid_index (i + 1, j + 1))
				grid [i][j].count += grid [i + 1][j + 1].ismine;
		} else
		{
			if (1 == valid_index (i, j - 2))
				grid [i][j].count += grid [i][j - 2].ismine;
			if (1 == valid_index (i, j + 2))
				grid [i][j].count += grid [i][j + 2].ismine;
			int odd = (j % 2);
			if (1 == valid_index (i - 1 + odd, j - 1))
				grid [i][j].count += grid [i - 1 + odd][j - 1].ismine;
			if (1 == valid_index (i + odd, j - 1))
				grid [i][j].count += grid [i + odd][j - 1].ismine;
			if (1 == valid_index (i - 1 + odd, j + 1))
				grid [i][j].count += grid [i - 1 + odd][j + 1].ismine;
			if (1 == valid_index (i + odd, j + 1))
				grid [i][j].count += grid [i + odd][j + 1].ismine;
		}
	}
	return (0);
}

//---------------------------------------------------------------------------
// main drawing function
//---------------------------------------------------------------------------
int draw_cell (int i, int j)
{
	int posx, posy, size;
	if (game_tile == TILE_HEX)
	{
		posx = i * HEX_SIZE * 1.5;
		if ((j % 2) == 1) posx += 3 * HEX_SIZE / 4;
		posy = j * HEX_SIZE / 2;
		size = HEX_SIZE - 2;

		if (grid [i][j].cleared == 0) setcolor (WHITE);
		else setcolor (DARKGRAY);
		moveto (posx + size / 4, posy + size);
		lineto (posx, posy + size / 2);
		lineto (posx + size / 4, posy);
		lineto (posx + 3 * size / 4, posy);
		setcolor (DARKGRAY);
		lineto (posx + size, posy + size / 2);
		lineto (posx + 3 * size / 4, posy + size);
		lineto (posx + size / 4, posy + size);
	} else
	{
		posx = i * REC_SIZE;
		posy = j * REC_SIZE;
		size = REC_SIZE - 2;

		if (grid [i][j].cleared == 0) setcolor (WHITE);
		else setcolor (DARKGRAY);
		moveto (posx, posy + size);
		lineto (posx, posy);
		lineto (posx + size, posy);
		setcolor (DARKGRAY);
		lineto (posx + size, posy  + size);
		lineto (posx, posy + size);
	}
	if (grid [i][j].marked == 1 && grid[i][j].cleared == 0)
	{
		print_text (posx + size / 2, posy + size / 2, "?", RED, 1);
	}
	if (grid [i][j].count > 0 && grid[i][j].cleared == 1)
	{
		char* str [9] = {"0", "1", "2", "3", "4", "5", "6", "7", "8" };
		print_text (posx + size / 2, posy + size / 2,
			str [grid[i][j].count], grid[i][j].count, 1);
	}
	if (game_over == 1 && grid[i][j].ismine == 1)
	{
		print_text (posx + size / 2, posy + size / 2, "*", RED, 1);
	}
	return (0);
}

//---------------------------------------------------------------------------
// draw the cell grid
//---------------------------------------------------------------------------
int draw_grid ()
{
	show_mouse ();
	limit_mouse ();
	hide_mouse ();
	setbkcolor (LIGHTGRAY);
	cleardevice ();

	for (int i = 0; i < grid_w; i ++)
	for (int j = 0; j < grid_h; j ++)
	{
		draw_cell (i, j);
	}
	setfillstyle (SOLID_FILL, LIGHTBLUE);
	bar (0, 460, 640, 480);
	print_text (270, 470, "MINESWEEPER, BC++3.1, 1998, TKXUYEN, R=RECTANGLE, H=HEXAGON, X=EXIT", RED, 1);

	show_mouse ();
	return (0);
}

//---------------------------------------------------------------------------
// mark a mine cell in grid
//---------------------------------------------------------------------------
int mark_grid (int x, int y)
{
	if (0 == valid_index (x, y)) return (1);
	if (grid [x][y].marked == 0) grid [x][y].marked = 1;
	else grid [x][y].marked = 0;
	draw_cell (x, y);
	return (0);
}

//---------------------------------------------------------------------------
// check_neighbours: check the 4-connective neighbour cells
//---------------------------------------------------------------------------
int check_neighbours (int i, int j)
{
	if (grid [i][j].cleared == 1) return (1);
	if (grid [i][j].ismine  == 0)
	{
		grid[i][j].cleared = 1;
		if (grid[i][j].count == 0)
		{
			if (game_tile == TILE_REC)
			{
				if (1 == valid_index (i - 1, j - 1))
					check_neighbours (i - 1, j - 1);
				if (1 == valid_index (i, j - 1))
					check_neighbours (i, j - 1);
				if (1 == valid_index (i + 1, j - 1))
					check_neighbours (i + 1, j - 1);
				if (1 == valid_index (i - 1, j))
					check_neighbours (i - 1, j);
				if (1 == valid_index (i + 1, j))
					check_neighbours (i + 1, j);
				if (1 == valid_index (i - 1, j + 1))
					check_neighbours (i - 1, j + 1);
				if (1 == valid_index (i, j + 1))
					check_neighbours (i, j + 1);
				if (1 == valid_index (i + 1, j + 1))
					check_neighbours (i + 1, j + 1);
			} else
			{
				if (1 == valid_index (i, j - 2))
					check_neighbours (i, j - 2);
				if (1 == valid_index (i, j + 2))
					check_neighbours (i,j + 2);
				int odd = (j % 2);
				if (1 == valid_index (i - 1 + odd, j - 1))
					check_neighbours (i - 1 + odd, j - 1);
				if (1 == valid_index (i + odd, j - 1))
					check_neighbours (i + odd, j - 1);
				if (1 == valid_index (i - 1 + odd, j + 1))
					check_neighbours (i - 1 + odd, j + 1);
				if (1 == valid_index (i + odd, j + 1))
					check_neighbours (i + odd, j + 1);

			}
		}
	}
	draw_cell (i, j);
	return (0);
}

//---------------------------------------------------------------------------
// check_grid: check a cell if it's mine or not
//---------------------------------------------------------------------------
int check_grid (int x, int y)
{
	if (x >= grid_w || y >= grid_h) return (1);
	if (grid [x][y].ismine == 1) 	// game over
	{
		game_over = 1;
		draw_grid ();
		print_text (320, 240, "!GAME OVER!", RED, 4);
		return (1);
	}
	check_neighbours (x, y);
	return (0);
}

//---------------------------------------------------------------------------
// check for a certain keystroke
//----------------------------------------------------------------------------
int check_key ()
{
	char ch = getch();
	switch (ch)
	{
		case 'x':
		case 'X':
			game_exit = 1;
			return (1);
		case 'r':
		case 'R':
			if (game_tile == TILE_HEX) game_tile = TILE_REC;
			break;
		case 'h':
		case 'H':
			if (game_tile == TILE_REC) game_tile = TILE_HEX;
			break;
		default: return (0);
	};
	game_over = 0; init_grid (); draw_grid ();
	return (1);
}

//---------------------------------------------------------------------------
// check if the game is win
//----------------------------------------------------------------------------
int check_win ()
{
	for (int i = 0; i < grid_w; i ++)
	for (int j = 0; j < grid_h; j ++)
	{
		if (grid[i][j].ismine == 1 && grid[i][j].marked == 0)
			return (0);
		if (grid[i][j].ismine == 0 && grid[i][j].cleared == 0)
			return (0);
	}
	return (1);
}


//---------------------------------------------------------------------------
// distance from mouse cursor to the given hexagon cell
//----------------------------------------------------------------------------
int distance_to_hex (int i, int j, int x, int y)
{
	int posx = i * (3 * HEX_SIZE / 2) + HEX_SIZE / 2;
	if ((j % 2) == 1) posx += 3 * HEX_SIZE / 4;
	int posy = j * HEX_SIZE / 2 + HEX_SIZE / 2;

	return ((x - posx) * (x - posx) + (y - posy) * (y - posy));
}

//---------------------------------------------------------------------------
// find the cell index under mouse position (x, y)
// index also returned in (x, y)
//----------------------------------------------------------------------------
int find_cell (int &x, int &y)
{
	if (game_tile == TILE_REC)
	{
		x /= REC_SIZE; y /= REC_SIZE;
		return (0);
	}
	int row2 = y * 2 / HEX_SIZE;
	int row1 = row2 - 1;
	int col2 = x * 2 / (HEX_SIZE * 3);
	int col1 = col2 - 1;
	int x1 = x, y1 = y;
	int dist = 32000, dist1;

	for (int i = col1; i <= col2; i++)
	for (int j = row1; j <= row2; j++)
	{
		if (0 == valid_index (i, j)) continue;
		dist1 = distance_to_hex (i, j, x1, y1);
		if (dist1 < dist)
		{
			dist = dist1;
			x = i; y = j;
		}
	}
	return (0);
}

//---------------------------------------------------------------------------
// main loop function
//---------------------------------------------------------------------------
int draw_game ()
{
	int x = 0, y = 0, click = 0, ret = 0;
	int lastx = 0, lasty = 0;

	while (game_exit == 0)
	{
		if (kbhit ()) check_key ();
		mouse_status (x, y, click);
		if (x != lastx || y != lasty)
		{
			char str [16];
			sprintf (str, "(%3dx%3d)", x, y);

			setfillstyle (SOLID_FILL, LIGHTBLUE);
			bar (540, 460, 640, 480);
			print_text (590, 470, str, MAGENTA, 1);
			lastx = x; lasty = y;
		}
		if (click > 0)
		{
			hide_mouse ();
			find_cell (x, y);
			switch (click)
			{
			case 1: ret = check_grid(x, y); break;
			case 3: break;
			case 2: ret = mark_grid (x, y); break;
			}
			show_mouse ();
			if (ret == 1) // game over
			{
				while (!kbhit()) if (1 == check_key()) break;
			}
		}
		if (1 == check_win ())
		{
			print_text (320, 240, "!CONGRATULATION!", RED, 4);
			while (!kbhit()) if (1 == check_key ()) break;
		}
		delay (100);
	}
	return (0);
}

//---------------------------------------------------------------------------
// main function
//---------------------------------------------------------------------------
void main ()
{
	init_game();
	draw_game ();
	term_game();
}
